"use strict";
/*
 * Unless explicitly stated otherwise all files in this repository are licensed
 * under the Apache 2.0 license (see LICENSE).
 * Copyright 2020 Datadog, Inc. for original work
 * Copyright 2021 GraphMetrics for modifications
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.LogarithmicMapping = void 0;
const helpers_1 = require("./helpers");
/**
 * A memory-optimal KeyMapping, i.e., given a targeted relative accuracy, it
 * requires the least number of keys to cover a given range of values. This is
 * done by logarithmically mapping floating-point values to integers.
 */
class LogarithmicMapping {
    constructor(relativeAccuracy) {
        this.relativeAccuracy = relativeAccuracy;
        this.multiplier =
            1 / Math.log1p((2 * relativeAccuracy) / (1 - relativeAccuracy));
        this.minIndexableValue = Math.max(Math.exp(helpers_1.MIN_INT_16 / this.multiplier + 1), (helpers_1.MIN_SAFE_FLOAT * (1 + relativeAccuracy)) / (1 - relativeAccuracy));
        this.maxIndexableValue = Math.min(Math.exp(helpers_1.MAX_INT_16 / this.multiplier - 1), Math.exp(helpers_1.EXP_OVERFLOW) / (1 + relativeAccuracy));
    }
    index(value) {
        const index = Math.log(value) * this.multiplier;
        if (index >= 0) {
            return ~~index;
        }
        else {
            return ~~index - 1; // faster than Math.Floor
        }
    }
    value(index) {
        return Math.exp(index / this.multiplier) * (1 + this.relativeAccuracy);
    }
    equals(other) {
        if (!(other instanceof LogarithmicMapping)) {
            return false;
        }
        return helpers_1.withinTolerance(this.multiplier, other.multiplier, 1e-12);
    }
}
exports.LogarithmicMapping = LogarithmicMapping;
