"use strict";
/*
 * Unless explicitly stated otherwise all files in this repository are licensed
 * under the Apache 2.0 license (see LICENSE).
 * Copyright 2020 Datadog, Inc. for original work
 * Copyright 2021 GraphMetrics for modifications
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.CollapsingHighestDenseStore = void 0;
const DenseStore_1 = require("./DenseStore");
const util_1 = require("./util");
/**
 * `CollapsingHighestDenseStore` is a dense store that keeps all the bins between
 * the bin for the `minKey` and the `maxKey`, but collapsing the left-most bins
 * if the number of bins exceeds `binLimit`
 */
class CollapsingHighestDenseStore extends DenseStore_1.DenseStore {
    /**
     * Initialize a new CollapsingHighestDenseStore
     *
     * @param binLimit The maximum number of bins
     * @param chunkSize The number of bins to add each time the bins grow (default 128)
     */
    constructor(binLimit, chunkSize) {
        super(chunkSize);
        this.binLimit = binLimit;
        this.isCollapsed = false;
    }
    /**
     * Merge the contents of the parameter `store` into this store
     *
     * @param store The store to merge into the caller store
     */
    merge(store) {
        if (store.count === 0) {
            return;
        }
        if (this.count === 0) {
            this.copy(store);
            return;
        }
        if (store.minKey < this.minKey || store.maxKey > this.maxKey) {
            this._extendRange(store.minKey, store.maxKey);
        }
        const collapseEndIndex = store.maxKey - store.offset + 1;
        let collapseStartIndex = Math.max(this.maxKey + 1, store.minKey) - store.offset;
        if (collapseEndIndex > collapseStartIndex) {
            const collapseCount = util_1.sumOfRange(store.bins, collapseStartIndex, collapseEndIndex);
            this.bins[this.length() - 1] += collapseCount;
        }
        else {
            collapseStartIndex = collapseEndIndex;
        }
        for (let key = store.minKey; key < collapseStartIndex + store.offset; key++) {
            this.bins[key - this.offset] += store.bins[key - store.offset];
        }
        this.count += store.count;
    }
    /**
     * Directly clone the contents of the parameter `store` into this store
     *
     * @param store The store to be copied into the caller store
     */
    copy(store) {
        super.copy(store);
        this.isCollapsed = store.isCollapsed;
    }
    _getNewLength(newMinKey, newMaxKey) {
        const desiredLength = newMaxKey - newMinKey + 1;
        return Math.min(this.chunkSize * Math.ceil(desiredLength / this.chunkSize), this.binLimit);
    }
    /**
     * Adjust the `bins`, the `offset`, the `minKey`, and the `maxKey`
     * without resizing the bins, in order to try to make it fit the specified range.
     * Collapse to the left if necessary
     */
    _adjust(newMinKey, newMaxKey) {
        if (newMaxKey - newMinKey + 1 > this.length()) {
            // The range of keys is too wide, the lowest bins need to be collapsed
            newMaxKey = newMinKey + this.length() + 1;
            if (newMaxKey <= this.minKey) {
                // Put everything in the first bin
                this.offset = newMinKey;
                this.maxKey = newMaxKey;
                this.bins.fill(0);
                this.bins[this.length() - 1] = this.count;
            }
            else {
                const shift = this.offset - newMinKey;
                if (shift > 0) {
                    const collapseStartIndex = newMaxKey - this.offset + 1;
                    const collapseEndIndex = this.maxKey - this.offset + 1;
                    const collapsedCount = util_1.sumOfRange(this.bins, collapseStartIndex, collapseEndIndex);
                    this.bins.fill(0, collapseStartIndex, collapseEndIndex);
                    this.bins[collapseStartIndex - 1] += collapsedCount;
                    this.maxKey = newMaxKey;
                    this._shiftBins(shift);
                }
                else {
                    this.maxKey = newMaxKey;
                    // Shift the buckets to make room for newMinKey
                    this._shiftBins(shift);
                }
                this.minKey = newMinKey;
                this.isCollapsed = true;
            }
        }
        else {
            this._centerBins(newMinKey, newMaxKey);
            this.minKey = newMinKey;
            this.maxKey = newMaxKey;
        }
    }
    /** Calculate the bin index for the key, extending the range if necessary */
    _getIndex(key) {
        if (key < this.minKey) {
            if (this.isCollapsed) {
                return this.length() - 1;
            }
            this._extendRange(key);
            if (this.isCollapsed) {
                return this.length() - 1;
            }
        }
        else if (key > this.maxKey) {
            this._extendRange(key);
        }
        return key - this.offset;
    }
}
exports.CollapsingHighestDenseStore = CollapsingHighestDenseStore;
